#pragma once
#include "RenderTechnique.h"

/**
	@brief Base abstract class of a caustic caster. 
	
	This technique defines that the given object needs a caustic photon map and a caustic cubemap.
	The caustic cube map will be used by caustic recievers.
*/
class CausticCasterRenderTechnique : virtual public RenderTechnique
{
public:
	/**
		@brief Constructor.

		@param startFrame					adds an offset to the current frame number to help evenly distribute updates between frames
		@param photonMapUpdateInterval		photon map and caustic cubemap update frequency
		@param photonMapResolution			photon map resolution
		@param causticCubeMapResolution		caustic cubemap resolution
		@param updateAllFace				defines if all cubemap faces should be updated in a frame or only one face per frame
		@param useDistance					tells if a distance cubemap impostor should be used in photon hit calculation (recommended)
		@param parentRenderable				the object to operate on
		@param parentTechniqueGroup			the TechniqueGroup this RenderedTechnique is attached to
	*/
	CausticCasterRenderTechnique( unsigned long startFrame,
							unsigned long photonMapUpdateInterval,
							unsigned int photonMapResolution,
							unsigned int causticCubeMapResolution,
							bool updateAllFace,
							bool useDistance,
							ElementaryRenderable* parentRenderable,
							TechniqueGroup* parentTechniqueGroup
							);
	~CausticCasterRenderTechnique();
	
	//inherited
	void runChanged(RenderingRunType runType, RenderingRun* run);
	//inherited
	virtual void runUpdated(RenderingRunType runType, RenderingRun* run);
	
protected:
	/**
			@brief defines if all cubemap faces should be updated in a frame or only one face per frame
	*/
	bool updateAllFace;							
	/**
			@brief tells if a distance cubemap impostor should be used in photon hit calculation (recommended)
	*/
	bool useDistance;
    
	/**
			@brief photonmap update frequency
	*/
	unsigned long photonMapUpdateInterval;
	/**
			@brief photonmap resolution
	*/
	unsigned int photonMapResolution;
	/**
			@brief caustic cubemap resolution
	*/
	unsigned int causticCubeMapResolution;
	/**
			@brief offset in frame number used during update
	*/
	unsigned long startFrame;	
	
	/**
			@brief Called if the changed run is a PhotonMapRenderingRun.

			@param run pointer to the changed PhotonMapRenderingRun
	*/		
	virtual void photonMapRunChanged(RenderingRun* run) = 0;
	/**
			@brief Called if the changed run is a CausticCubeMapRenderingRun.

			@param run pointer to the changed CausticCubeMapRenderingRun
	*/		
	virtual void causticCubeMapRunChanged(RenderingRun* run) = 0;
	/**
			@brief Called if the changed run is a DistanceCubeMapRenderingRun.

			@param run pointer to the changed DistanceCubeMapRenderingRun
	*/	
	virtual void distanceCubeMapRunChanged(RenderingRun* run) = 0;
	/**
			@brief Called if the updated run is a DistanceCubeMapRenderingRun.

			@param run pointer to the updated DistanceCubeMapRenderingRun
	*/	
	virtual void distanceCubeMapRunUpdated(RenderingRun* run) = 0;
	/**
			@brief Creates a PhotonMapRenderingRun.

			@return the new PhotonMapRenderingRun instance.
	*/
	virtual RenderingRun* createPhotonMapRun() = 0;
	/**
			@brief Creates a CausticCubeMapRenderingRun.

			@return the new CausticCubeMapRenderingRun instance.
	*/
	virtual RenderingRun* createCausticCubeMapRun() = 0;
	/**
			@brief Creates a DistanceCubeMapRenderingRun.

			@return the new DistanceCubeMapRenderingRun instance.
	*/
	virtual RenderingRun* createDistanceCubeMapRun() = 0;
};